<?php

/*
 * définititon des actions à distance sur les postes
 *
 * * @Projet LCS / SambaEdu
 *
 * @Auteurs Equipe Sambaedu
 *
 * @Version
 *
 * @Note: AD :
 * - netbootinitialization => id
 * - netbootguid => token
 *
 * les actions en cours sont enregistrées dans un tableau caché par apcu.
 *
 * action : array(id_xx => array( 'type' , machines => array('token', 'action'), 'statut', 'progress' )))
 *
 * @Licence Distribue sous la licence GPL
 *
 */

/**
 * Corrige si besoin les comptes utilisés pour les actions automatisées
 *
 * @param unknown $config
 */
function check_system_accounts($config)
{

    // utilisateur se4install pour les install auto
    if (empty($config['se4install_name'])) {
        set_config('se4install_name', "se4install", "clients");
    }

    if (empty($config['se4install_passwd'])) {
        set_config('se4install_passwd', create_random_password(8, false), "clients");
    }
    if (! userexist($config, $config['se4install_name']))
        create_user($config, $config['se4install_name'], "install", "auto", $config['se4install_passwd'], "", "", "");
    if (count(search_ad($config, $config['se4install_name'], "memberof", "Domain Admins")) == 0)
        groupaddmember($config, $config['se4install_name'], "Domain Admins");
    if (! user_valid_passwd($config, $config['se4install_name'], $config['se4install_passwd']))
        usersetpassword($config, $config['se4install_name'], $config['se4install_passwd']);
    // suppression adminse qui ne doit pas être au domaine
    if (! empty($config['adminse_name']) && userexist($config, $config['adminse_name']) && ! in_array($config['adminse_name'], [
        "admin",
        "administrator",
        $config['se4install_name']
    ]))
        userdel($config, $config['adminse_name']);
}

/**
 * récupére les infos des disques du modèle, envoyé par lsblkid
 * par defaut toutes les partitions sont clonées
 * le statut contient l'avancement du clonage (0-100), -1 en cas d'erreur
 *
 * @param
 *            $config
 * @param $partition :
 *            sortie de lsblkid
 * @return array[]|number
 */
function get_disks($config, $partitions)
{
    $partitions = explode("\n", $partitions);
    $disks = array();
    $index = - 1;
    foreach ($partitions as $ligne) {
        $match = array();
        if (preg_match("@^(/dev/sd[a-z]|/dev/nvme[0-9]n[0-9])\s*$@", $ligne, $match)) {
            $index ++;
            $disks[] = array(
                'device' => $match[1],
                'clone' => false
            );
            $indexp = 0;
        } elseif (preg_match("@^(/dev/sd[a-z][0-9]+|/dev/nvme[0-9]n[0-9]p[0-9]+)\s+(.*)\s*$@", $ligne, $match)) {
            $disks[$index]['partitions'][$indexp ++] = array(
                'device' => $match[1],
                'fs' => $match[2],
                'clone' => true
            );
            $disks[$index]['clone'] = true;
            /*
             * $disks[$match[1]][$match[2]]['statut'] = 0; //pour info
             * $disks[$match[1]][$match[2]]['start'] = 0; //pour info
             * $disks[$match[1]][$match[2]]['size'] = 0; //pour info
             * $disks[$match[1]][$match[2]]['name'] = ""; //pour info
             * $disks[$match[1]][$match[2]]['type'] = ""; //pour info
             * $disks[$match[1]][$match[2]]['uuid'] = ""; //pour info
             *
             */
        }
    }
    return $disks;
}

/**
 * fixe les disques et partitions à cloner
 * si le disque n'est pas choisi aucune partition ne sera clonée
 *
 * @param unknown $id
 * @param unknown $partitions
 */
function set_disks($id, $partitions)
{
    lock('actions');
    $actions = fetch_action();

    foreach ($actions[$id]['disks'] as $d => $disk) {
        if (in_array($disk['device'], $partitions)) {
            $actions[$id]['disks'][$d]['clone'] = true;
            foreach ($disk['partitions'] as $p => $partition) {
                if (in_array($partition['device'], $partitions)) {
                    $actions[$id]['disks'][$d]['partitions'][$p]['clone'] = true;
                } else {
                    $actions[$id]['disks'][$d]['partitions'][$p]['clone'] = false;
                }
            }
        } else {
            $actions[$id]['disks'][$d]['clone'] = false;
            foreach ($disk['partitions'] as $p => $partition) {
                $actions[$id]['disks'][$d][$p]['clone'] = false;
            }
        }
    }
    // unset($actions[$id]['disks'][0][0]);
    apcu_store('actions', $actions);
    unlock('actions');
}

function set_disk_size(&$disks, $sfdisk)
{
    $partitions = explode("\n", $sfdisk);
    foreach ($partitions as $ligne) {
        $match = array();
        if (preg_match("@^(/dev/sd[a-z][0-9]|n0p[0-9]+|/dev/nvme[0-9]n[0-9]p[0-9]+)\s+.*size=\s+([0-9]+).*$@", $ligne, $match)) {
            foreach ($disks as $index => $disk) {
                foreach ($disk['partitions'] as $indexp => $partition) {
                    if ($partition['device'] == $match[1]) {
                        $disks[$index]['partitions'][$indexp]['size'] = (int) $match[2] / 2;
                        break;
                    }
                }
            }
        }
    }
}

/**
 * enregistre l'avanceemnt du clonage à partir des infos envoyées par partclone
 *
 * @param string $id
 * @param array $progress
 */
function set_progress($id, $progress)
{
    $match = array();
    if (preg_match("/([0-9.]+)%/", $progress, $match)) {
        lock('actions');
        $actions = fetch_action();
        if (isset($actions[$id])) {
            if ($actions[$id]['type'] == "clonage") {
                foreach ($actions[$id] as $membre) {
                    if (! empty($membre['role'])) {
                        if ($membre['role'] == 'modele') {
                            $etape = $membre['ret'];
                        }
                    }
                }
                if (isset($actions[$id]['disks']) && is_array($actions[$id]['disks'])) {
                    $disks = $actions[$id]['disks'];
                    $total = get_total_size($disks);
                    // if ($total == 0) {
                    // $fdisk = [];
                    // $e = 0;
                    // foreach ($disks as $disk) {
                    // if ($disk['clone']) {
                    // $fdisk = array_merge($fdisk, file_get_contents(sys_get_temp_dir() . "/" . md5($id) . "-" . $e . ".txt"));
                    // $e ++;
                    // }
                    // }
                    // set_disk_size($disks, $fdisk);
                    // $actions[$id]['disks'] = $disks;
                    // $total = get_total_size($disks);
                    // }
                    $devs = get_clone_parts($disks);
                    $fait = round(get_total_size($disks, $etape - 1) + $devs[$etape]['size'] * ($match[1] / 100 - 1), 1);
                    if ($fait > $actions[$id]['progress']['fait']) {
                        $speed = round(($fait - $actions[$id]['progress']['fait']) / (time() - $actions[$id]['progress']['time']) / 512, 1);
                        $time = time();
                    } else {
                        $speed = $actions[$id]['progress']['speed'];
                        $time = $actions[$id]['progress']['time'];
                    }
                } else {
                    $total = 100;
                    $fait = 0;
                }
            } else {
                $total = 100;
                $fait = $match[1];
            }
            $actions[$id]['progress'] = array(
                'total' => $total,
                'fait' => $fait,
                'avancement' => round($fait / $total * 100, 1),
                'time' => $time ?? time(),
                'speed' => $speed ?? 0
            );
            apcu_store('actions', $actions);
        }
        unlock('actions');
    }
}

function get_clone_disks($disks)
{
    $list = array();
    foreach ($disks as $disk) {
        if ($disk['clone'])
            $list[] = $disk['device'];
    }
    return $list;
}

function get_clone_parts($disks)
{
    $list = array();
    foreach ($disks as $disk) {
        if ($disk['clone'])
            foreach ($disk['partitions'] as $part)
                if ($part['clone'])
                    $list[] = array(
                        'device' => $part['device'],
                        'fs' => $part['fs'],
                        'size' => $part['size']
                    );
    }
    return $list;
}

function get_total_size($disks, $index = "")
{
    $size = $i = 0;
    foreach ($disks as $disk) {
        if ($disk['clone'])
            foreach ($disk['partitions'] as $part)
                if ($part['clone']) {
                    if (isset($part['size'])) {
                        $size = $size + $part['size'];
                    }
                    if (! empty($index) && ++ $i == $index)
                        return $size;
                }
    }
    return $size;
}

/**
 * incremente le port udp utilisé par le clonage, de façon à etre unique et pair
 *
 * @param
 *            string
 * @return number
 */
function get_clone_port($id)
{
    $port = 9000;
    $ports = array();
    lock('actions');
    $actions = fetch_action();
    if (! empty($actions[$id]['port'])) {
        $port = $actions[$id]['port'];
    } else {
        foreach ($actions as $action) {
            if (! empty($action['port']))
                $ports[] = $action['port'];
        }
        while (in_array($port, $ports)) {
            $port = $port + 2;
        }
    }
    $actions[$id]['port'] = $port;
    apcu_store('actions', $actions);
    unlock('actions');

    return $port;
}

function get_action_member($config, $id, $role)
{
    $list = array();
    $action = fetch_action()[$id] ?? [];
    foreach ($action as $uuid => $membre) {
        if (! empty($membre['role'])) {
            if ($membre['role'] == $role) {
                $list[] = search_machine($config, $uuid)['cn'];
            }
        }
    }
    return $list;
}

function set_statut($id, $statut = "")
{
    lock('actions');
    $actions = fetch_action();
    $script = $actions[$id]['statut'] ?? "inconnu";
    if (isset($actions[$id])) {
        foreach ($actions[$id] as $membre) {
            if (! empty($membre['role'])) {
                if ($membre['role'] == 'modele') {
                    // $script = $membre['script'];
                    $script = $membre['etape'] ?? "attente...";
                }
            }
        }
        if (($actions[$id]['type'] != "clonage" || $etape == "emit-partitions" || $etape == "change-hostname") && ! empty($statut)) {
            $script = $statut;
        }
        $actions[$id]['statut'] = $script;
        apcu_store('actions', $actions);
    }
    unlock('actions');
    return $script;
}

function wait_clone($id)
{
    $pret = true;
    $nb = $p = 0;
    $actions = fetch_action()[$id] ?? [];
    foreach ($actions as $membre) {
        if (! empty($membre['role'])) {
            if ($membre['role'] == 'modele') {
                $etape = $membre['ret'] + 1;
            }
        }
    }
    foreach ($actions as $membre) {
        if (! empty($membre['role'])) {
            if ($membre['role'] == 'clone') {
                $nb ++;
                $mp = (($membre['etape'] == "receive-partitions") && ($membre['ret'] == $etape));
                if ($mp)
                    $p ++;
                $pret = ($pret && $mp);
            }
        }
    }
    // il faut qu'au moins un clone ait été défini, sinon on attend
    $pret = $pret && ($nb > 0);
    return array(
        'pret' => $pret,
        'clones' => "(" . $p . "/" . $nb . ")"
    );
}

function toggle_action_keep($id)
{
    lock('actions');
    $actions = fetch_action();
    if (isset($actions[$id]['keep'])) {
        if ($actions[$id]['keep'])
            $actions[$id]['keep'] = false;
        else
            $actions[$id]['keep'] = true;
    } else {
        $actions[$id]['keep'] = true;
    }
    apcu_store('actions', $actions);
    unlock('actions');
}

function refresh_page($form)
{
    $stylecss = "/elements/style_sheets/sambaedu.css";
    $html = "<meta http-equiv='Content-Type' content='text/html; charset=UTF-8' />
<link type='text/css' rel='stylesheet' href='$stylecss' />
<div id='decompte' style='float: right; border: 1px solid black;'></div>
<script type='text/javascript'>
    cpt=30;
    compte_a_rebours='y';
    
    
    /**
     * Decompte le temps
     * @language Javascript
     * @Parametres
     * @return le decompte qui s'affiche
     */
     
     
    function decompte(cpt){
        if(compte_a_rebours=='y'){
            document.getElementById('decompte').innerHTML=cpt;
            if(cpt>0){
                cpt--;
            }
            else{
                    document." . $form . ".submit()
           }
                        
            setTimeout(\"decompte(\"+cpt+\")\",1000);
    	}
	   else{
		  document.getElementById('decompte').style.display='none';
       }
    }
    decompte(cpt);
</script>\n";
    return $html;
}

function form_actions_statut($type = "")
{
    $html = "<TABLE border=1 width=\"80%\">\n";
    $html .= "<th colspan=5 class=menuheader style=\"height: 30\">Liste des actions</th>\n";

    $html .= "<tr><td class='menuheader' align=\"center\">Identifiant</td>";
    $html .= "<td class='menuheader' align=\"center\">Type</td>";
    $html .= "<td class='menuheader' align=\"center\">Avancement</td>";
    $html .= "<td class='menuheader' align=\"center\">Statut</td>";
    $html .= "<td class='menuheader' align=\"center\">Selectionner</td>";
    $html .= "</tr>\n";
    $actions = fetch_action();

    foreach ($actions as $id => $action) {
        if (($id != "default") && (empty($type) || $type == $action['type'])) {
            $statut = set_statut($id);
            $progress = $action['progress']['avancement'] ?? 0;
            $html .= "<tr><td>" . $id . "</td>";
            $html .= "<td>" . $action['type'] . "</td>";
            // $html .= "<td><label for=\"" . $id . "\">fait : " . $action['progress']['avancement'] . "%</label>";
            // $html .= "<progress id=\"" . $id . "\" max=\"100\" value=" . $action['progress']['avancement'] . "> " . $action['progress']['avancement'] . "% </progress></td>";
            $html .= "<td><progress  max=\"100\" value=" . $progress . "> " . $progress . "% </progress>  " . $progress . " %</td>";
            $html .= "<td>" . $statut . "</td>";
            $html .= "<td align=\"center\"><input type=\"checkbox\" name=\"id[]\" value=\"" . $id . "\"></td>";
            $html .= "</tr>\n";
        }
    }
    $html .= "</TABLE>";
    // $html .= form_action_type($type);
    return $html;
}

function form_action_type($type = "voir")
{
    $types = array(
        "voir" => "Voir l'action",
        "supprimer" => "Supprimer l'action",
        "clonage" => "Clonage en direct avec Sysrescd",
        "install" => "Installation automatique des postes",
        "renomme" => "Renommage d'un poste",
        // "renomme_sysprep" => "Renommage d'un poste (sysprep)",
        "post" => "Post-int&#233;gration ou migration se3",
        "wpkg" => "Forcer wpkg en mode graphique",
        "reboot" => "Allumer ou rebooter"
    );
    if (! empty($config['ltsp_ip'])) {
        $types = array_merge($types, [
            'ltsp' => "Booter un client LTSP"
        ]);
    }
    asort($types);
    // Commenté pour le moment car aucune prise en compte dans actions.inc.php et donc ne fonctionne pas
    // $clonezilla_type = array("image" => "Sauvegarde et restauration d'une image Clonezilla");
    // $types = array_merge($types,$clonezilla_type);
    $html = "Choix du type d'actions ";
    $html .= "<SELECT NAME=\"type\" onchange=\"this.form.submit()\">";
    $html .= "<option value=\"\">Rafraichir</option>";
    foreach ($types as $t => $txt) {
        $html .= "<option value=\"$t\"";
        if ($type == $t) {
            $html .= " selected>";
        } else {
            $html .= ">";
        }
        $html .= "$txt</option>";
    }
    $html .= "</SELECT>";
    $html .= " <input type=\"submit\" value=\"" . gettext("Valider") . "\">\n";
    return $html;
}

// @TODO differencier le choix en fonction de l'étape en cours
function form_action_clonage($action = "", $current = [])
{
    // début
    $actions = array(
        'new' => "Nouveau clonage ou modèle "
    );
    // modele choisi
    $actions = array(
        'prepare' => "Préparer le modèle",
        'add' => "Ajouter des clones",
        'remove' => "Supprimer des clones",
        'start' => "(Re)Démarrer les clones",
        'del' => "Supprimer totalement",
        'submit' => "Valider/Actualiser"
    );
    // modele pret
    $actions = array(
        'keep' => "Conserver le modèle pour un clonage ultérieur",
        'add' => "Ajouter des clones",
        'remove' => "Supprimer des clones",
        'start' => "(Re)Démarrer les clones",
        'disk' => "Choisir les partitions à cloner",
        'del' => "Supprimer totalement",
        'submit' => "Valider/Actualiser"
    );

    $actions = array(
        'submit' => "Valider/Actualiser",
        'start' => "Démarrer",
        'prepare' => "Préparer le modèle",
        'keep' => "Conserver le modèle pour un clonage ultérieur",
        'reset' => "Redémarrer le clonage (sans refaire le modèle)",
        'new' => "Nouveau clonage ou modèle ",
        'add' => "Ajouter des clones",
        'remove' => "Supprimer des clones",
        'disk' => "Choisir les partitions à cloner",
        'del' => "Supprimer totalement"
    );
    $html = "<SELECT NAME=\"action\" id=\"action_clonage\" onchange=\"this.form.submit()\">";
    $html .= "<option value=\"\">Choisir l'action...</option>";
    asort($actions);
    foreach ($actions as $a => $txt) {
        $html .= "<option value=\"$a\"";
        if ($action == $a) {
            $html .= " selected>";
        } else {
            $html .= ">";
        }
        $html .= "$txt</option>";
    }
    $html .= "</SELECT>";
    $html .= " <input type=\"submit\" value=\"" . gettext("Valider") . "\">\n";

    return $html;
}

function form_choix_install($action = "")
{
    if (file_exists("/var/sambaedu/unattended/install/os/Win10/sources/boot.wim")) {
        $actions['wimboot10'] = "Installation auto Windows 10";
        $actions['wimboot10disk'] = "Installation auto Windows 10 avec partitionnment personnalisé";
        $actions['wimboot10perso'] = "Installation auto Windows 10 (hors domaine)";
    }
    if (file_exists("/var/sambaedu/unattended/install/os/Win11/sources/boot.wim")) {
        $actions['wimboot11'] = "Installation auto Windows 11";
    }
    if (file_exists("/var/sambaedu/unattended/install/os/debian-installer")) {
        $actions = array_merge($actions, array(
            'deb_mate' => "Installation auto Debian Mate",
            'deb_gnome' => "Installation auto Debian Gnome",
            'deb_cinnamon' => "Installation auto Debian Cinnamon",
            'deb_kde' => "Installation auto Debian KDE",
            'deb_base' => "Installation auto Debian terminal",
            'deb_gnome_perso' => "Installation auto Debian Gnome (hors domaine)"
        ));
    }
    if (file_exists("/var/sambaedu/unattended/install/os/ubuntu-installer")) {
        $actions = array_merge($actions, array(
            'ubuntu64' => "Installation auto Ubuntu (hors domaine)"
        ));
    }
    asort($actions);
    $html = "Choisir l'OS à installer\n";
    $html .= "<SELECT NAME=\"action\" onchange=\"submit()\">";
    $html .= "<option value=\"\">Installation...</option>";
    foreach ($actions as $a => $txt) {
        $html .= "<option value=\"$a\"";
        if ($action == $a) {
            $html .= " selected>";
        } else {
            $html .= ">";
        }
        $html .= "$txt</option>";
    }
    $html .= "</SELECT>";
    return $html;
}

function form_choix_role($role = "")
{
    $html = "choisir le rôle de la machine pour le clonage\n";
    $html .= "<SELECT NAME=\"role\" onchange=\"submit()\">";
    $html .= "<option value=\"\">Role...</option>";
    $roles = array(
        "modele",
        "clone"
    );
    foreach ($roles as $r) {
        $html .= "<option value=\"$r\"";
        if ($role == $r) {
            $html .= " selected>";
        } else {
            $html .= ">";
        }
        $html .= "$r</option>";
    }
    $html .= "</SELECT>";
    return $html;
}

function form_choix_serveur_ltsp($config, $script = "")
{
    $html = "choisir le serveur ltsp à utiliser\n";
    $html .= "<SELECT NAME=\"script\">";
    $html .= "<option value=\"\">choisir...</option>";
    groupadd($config, "serveurs_ltsp", $config['parcs_rdn'], "Serveurs de terminaux réseau LTSP");
    foreach (list_members_parc($config, "serveurs_ltsp", false) as $s) {
        $html .= "<option value=\"$s\"";
        if ($s == $script) {
            $html .= " selected>";
        } else {
            $html .= ">";
        }
        $html .= "$s</option>";
    }
    $html .= "</SELECT>";
    return $html;
}

function form_choix_reservation($action)
{
    $html = "Action dhcp :\n";
    $html .= "<SELECT NAME=\"action\" onchange=\"submit()\">";
    $html .= "<option value=\"\">choisir...</option>";
    $roles = array(
        "reservation",
        "dynamique"
    );
    foreach ($roles as $r) {
        $html .= "<option value=\"$r\"";
        if ($action == $r) {
            $html .= " selected>";
        } else {
            $html .= ">";
        }
        $html .= "$r</option>";
    }
    $html .= "</SELECT>";
    return $html;
}

function form_new_name($name, $clones)
{
    if (count($clones) == 1) {
        $html = "Choisir le nouveau nom pour la machine " . $clones[0] . " :\n";
        $html .= " <INPUT TYPE=\"text\" NAME=\"modele\" VALUE=\"" . $clones[0] . "\" SIZE=\"15\" >"; // onchange=submit()>";
    } else {
        $html = "Choisir le préfixe pour le nouveau nom des machines. Toutes les machines choisies pour être renommées auront cepréfixe de nom suivi de deux chiffres \n";
        $html .= " <INPUT TYPE=\"text\" NAME=\"modele\" VALUE=\"" . $name . "\" SIZE=\"15\" >-XX "; // onchange=submit()>";
    }
    foreach ($clones as $clone) {
        echo form_hidden("clones[]", $clone);
    }
    // $html .= "<input type=\"hidden\" name=\"name\" value=\"" . $name . "\">\n";
    $html .= "<input type=\"submit\" value=\"" . gettext("Renommer") . "\">\n";
    return $html;
}

function form_action_membres($config, $id)
{
    $action = fetch_action()[$id] ?? [];
    $progress = $action['progress']['avancement'] ?? 0;
    $speed = $action['progress']['speed'] ?? "0";
    $html = "<TABLE border=1 width=\"80%\">\n";
    $html .= "<th colspan=4 class=menuheader style=\"height: 30\">Liste des actions</th>\n";

    $html .= "<tr><td class='menuheader' align=\"center\">Identifiant</td>";
    $html .= "<td class='menuheader' align=\"center\">Type</td>";
    $html .= "<td class='menuheader' align=\"center\">Avancement</td>";
    $html .= "<td class='menuheader' align=\"center\">Statut</td>";
    $html .= "</tr>\n";
    $statut = $action['statut'] ?? "";
    $html .= "<tr><td>" . $id . "</td>";
    $html .= "<td>" . @$action['type'] . "</td>";
    $html .= "<td align=\"center\"> <progress  max=\"100\" value=" . $progress . "> " . $progress . "% </progress>  " . $progress . " % Débit = " . $speed . " Mb/s</td>";
    $html .= "<td>" . $statut . "</td>";
    $html .= "</tr>\n";
    $html .= "</TABLE>";

    $html .= "<TABLE border=1 width=\"80%\">\n";
    $html .= "<th colspan=7 class=menuheader style=\"height: 30\">Liste des postes</th>\n";
    $html .= "<tr><td class='menuheader' align=\"center\">Role</td>";
    $html .= "<td class='menuheader' align=\"center\">Nom</td>";
    $html .= "<td class='menuheader' align=\"center\">Ip</td>";
    $html .= "<td class='menuheader' align=\"center\">Mac</td>";
    $html .= "<td class='menuheader' align=\"center\">script</td>";
    $html .= "<td class='menuheader' align=\"center\">étape</td>";
    $html .= "<td class='menuheader' align=\"center\">Selectionner</td>";
    $html .= "</tr>\n";

    foreach ($action as $uuid => $membre) {
        if (! empty($membre['role'])) {
            $machine = get_action($config, $uuid);
            $etape =  $membre['etape'] ?? "attente...";
            $html .= "<tr><td>" . $membre['role'] . "</td>";
            $html .= "<td>" . $machine['cn'] . "</td>";
            $html .= "<td>" . @$machine['iphostnumber'] . "</td>";
            $html .= "<td>" . $machine['networkaddress'] . "</td>";
            $html .= "<td>" . $membre['script'] . "</td>";
            $html .= "<td>" . $etape . "</td>";
            if ($membre['role'] == "modele") {
                if (@$action['keep'])
                    $html .= "<td align=\"center\">Conserver</td>";
                else
                    $html .= "<td></td>";
            } else {
                $html .= "<td align=\"center\"><input type=\"checkbox\" name=\"clones[]\" value=\"" . $uuid . "\"></td>";
            }
            $html .= "</tr>\n";
        }
    }
    $html .= "</TABLE>";
    $html .= "<br>";
    $html .= "<input type=\"hidden\" name=\"id\" value=\"" . $id . "\">\n";

    return $html;
}

function form_choix_poste($config, $action, $filtre, $choix, $role, $id, $parc = "")
{
    $html = "Choix des postes pour l'action $role :<p>";
    $list_parcs = search_parcs($config, "*");
    if (count($list_parcs) > 0) {
        $html .= "Choisir un parc (facultatif) : <SELECT NAME=\"parc\" SIZE=\"1\" onchange=\"this.form.submit()\">";
        $html .= "<option value=\"\">Sélectionner</option>";
        for ($loop = 0; $loop < count($list_parcs); $loop ++) {
            $html .= "<option value=\"" . $list_parcs[$loop]["name"] . "\"";
            if ($parc == $list_parcs[$loop]["name"]) {
                $html .= " selected";
            }
            $html .= ">" . $list_parcs[$loop]["name"] . "\n";
            $html .= "</option>";
        }
        $html .= "</SELECT><p>\n";
    }
    $html .= "Filtre sur le nom des machines (laisser vide pour toutes)  : <INPUT TYPE=\"text\" name=\"filtre\" VALUE=\"$filtre\" SIZE=\"14\">";
    $html .= "<input type=\"hidden\" name=\"action\" value=\"$action\">\n";
    $html .= "<input type=\"hidden\" name=\"role\" value=\"$role\">\n";
    $html .= "<input type=\"hidden\" name=\"id\" value=\"$id\">\n";
    $html .= "<input type=\"hidden\" name=\"choix\" value=\"1\">\n";
    $html .= "<br>";
    if (empty($filtre))
        $filtre = "*";
    else
        $filtre = "*" . $filtre . "*";
    if ($choix == 1) {
        $list_computer = array_column(search_ad($config, $filtre, "machine", $config['dn']['computers']), 'cn');
        if (! empty($parc)) {
            $mps = list_members_parc($config, $parc);
            $list_computer = array_uintersect($mps, $list_computer, "strcasecmp");
        }
        if (count($list_computer) > 15)
            $size = 15;
        else
            $size = count($list_computer);
        if (count($list_computer) == 1) {
            $html .= gettext("Nom ou adresse IP de la machine : ");
            if ($role == "clone" || $role == "wpkg" || $role == "install" || $role == "renomme" || $role == "ltsp")
                $html .= "<input type=hidden  name=\"clones[0]\" value=" . $list_computer[0] . ">" . $list_computer[0] . "\n";
            // elseif ($role == "install")
            // $html .= "<p><select size=\"" . $size . "\" name=\"name\">\n";
            else
                $ma = get_action($config, $list_computer[0]);
            if (@$ma['action']['keep'] && @$ma['action']['role'] == "modele") {
                if ($ma['action']['role'] == "modele")
                    $html .= "<input type=hidden  name=\"modele\" value=" . $list_computer[0] . ">" . $list_computer[0] . "\n";
            } else {
                $html .= "<input type=hidden  name=\"modele\" value=" . $list_computer[0] . ">" . $list_computer[0] . "\n";
            }
        } elseif (count($list_computer) > 1) {
            sort($list_computer);
            $html .= gettext("Nom ou adresse IP de la machine : ");
            if ($role == "clone" || $role == "wpkg" || $role == "install" || $role == "renomme" || $role == "ltsp")
                $html .= "<p><select size=\"" . $size . "\" name=\"clones[]\" multiple>\n";
            // elseif ($role == "install")
            // $html .= "<p><select size=\"" . $size . "\" name=\"name\">\n";
            else
                $html .= "<p><select size=\"" . $size . "\" name=\"modele\">\n";
            foreach ($list_computer as $machine) {
                $ma = get_action($config, $machine);
                if (@$ma['action']['keep'] && @$ma['action']['role'] == "modele") {
                    if ($ma['action']['role'] == "modele")
                        $html .= "<option disabled value=\"" . $machine . "\">" . $machine . " (préparée)</option>\n";
                } else {
                    $html .= "<option value=\"" . $machine . "\">" . $machine . "</option>\n";
                }
            }
            $html .= "</select>\n";
            $html .= "<br>";
        }
    }
    $html .= " <input type=\"submit\" value=\"" . gettext("Valider") . "\">\n";
    $html .= "<u onmouseover=\"return escape" . gettext("('Donner le début du nom ou de l\'adresse IP de la machine que vous souhaitez trouver.')") . "\"><img name=\"action_image2\"  src=\"../elements/images/system-help.png\" alt=\"help\"></u> ";
    return $html;
}

function form_hidden($name, $val)
{
    return "<input type=\"hidden\" name=\"" . $name . "\" value=\"" . $val . "\">\n";
}

function form_choix_partitions($id)
{
    $action = fetch_action()[$id];
    $html = "<SELECT NAME=\"partitions[]\" SIZE=\"5\" MULTIPLE=\"true\">";
    $html .= "<option value=\"\">Selectionner les disques et partitions...</option>";
    foreach ($action['disks'] as $disk) {
        $html .= "<option value=\"" . $disk['device'] . "\"";
        if ($disk['clone'])
            $html .= " selected";
        $html .= ">Copier le MBR de " . $disk['device'] . "</option>";
        foreach ($disk['partitions'] as $partition) {
            $html .= "<option value=\"" . $partition['device'] . "\"";
            if ($partition['clone'])
                $html .= " selected";
            $html .= "> Cloner " . $partition['device'] . " (" . $partition['fs'] . ", " . round($partition['size'] / (1024 * 1024), 1) . " G)</option>";
        }
    }
    $html .= "</SELECT>";
    $html .= "<input type=\"hidden\" name=\"id\" value=\"" . $id . "\">\n";
    return $html;
}

function form_action_id($type = "")
{
    $html = "<FORM action=\"lance_action.php\" method=\"post\">\n";
    "<SELECT NAME=\"id\" SIZE=\"20\">";
    if ($actions = fetch_action()) {
        foreach ($actions as $id => $action) {
            if (empty($type) || ($action['type'] == $type)) {
                $html .= "<option value=\"" . $id . "\"> Id : " . $id . " (" . $action['type'] . ")</option>";
            }
        }
    }
    $html .= "</SELECT>";
    $html .= "<input type=\"submit\" value=\"" . gettext("Valider") . "\">\n";
    $html .= "</FORM>\n";
    return $html;
}

function select_actions($actions, $module = "")
{
    $option = "";
    foreach ($actions as $action) {
        if (empty($module) || ($action['module'] == $module)) {
            $option .= "<option value=\"" . $action['nom'] . "\">" . $action['description'] . "</option>";
        }
    }
    return $option;
}

/**
 * récupère le tableau action
 *
 * @return string[][]|string[][][]|number[][][]
 */
function fetch_action()
{
    if (! ($actions = apcu_fetch('actions'))) {
        $actions = array(
            'default' => array(
                'type' => 'default',
                'default' => array(
                    'role' => '',
                    'script' => "default",
                    'ret' => - 1
                )
            )
        );
        apcu_add('actions', $actions);
    }
    return $actions;
}

/**
 * Récupère l'action programmée pour la machine, au boot ipxe, ou lors d'actions à distance
 * on récupère l'action et un token unique commun à une action et qui permet d'authentifier la requête
 * si l'enregistrement ad n'existe pas, on renvoie une action avec l'@mac comme id
 * Cette fonction ne met rien en cache.
 * Elle n'ecrit pas dans l'Ad sauf si mac est passée et qu'il manque uuid dans l'ad (migration)
 *
 * @param array $config
 * @param string $uuid
 *            : uuid ip, ou nom de la machine
 * @param string $mac
 *            : mac ip, ou nom de la machine
 * @return array : tableau des attributes de la machine, avec les valeurs par défaut pour 'netbootinitialization' (action)
 *         et 'netbootguid' (token)
 */
function get_action(array $config, string $uuid = "", string $mac = "")
{
    if (! empty($uuid)) {
        $machine = search_machine($config, $uuid, true, true);
    } else {
        $machine = [];
    }
    if (count($machine) == 0 && ! empty($mac)) {
        // pas trouvé l'uuid, on essaie avec le 2ème argument (mac, ou autre)
        $machine = search_machine($config, $mac, true, true);
    }
    if (count($machine) > 0 && ! empty($uuid) && $machine['cn'] != strtolower($uuid) && isset($machine['networkaddress']) && $machine['networkaddress'] != $uuid) {
        // on enregistre l'uuid si c'est bien un uuid
        $machine = register_machine_hardware($config, $machine, $uuid);
    }

    // recherche des actions en cours pour la machine
    $actions = fetch_action();
    if (count($machine) > 0) {
        // un enregistrement AD existe, on l'utilise
        $uuid = $machine['netbootguid'];
        $machine['netbootinitialization'] = $machine['netbootinitialization'] ?? "";
        if (isset($actions[$machine['netbootinitialization']][$machine['netbootguid']])) {
            $machine['action'] = $actions[$machine['netbootinitialization']];
        } else {
            foreach ($actions as $id => $action) {
                if (! empty($machine['netbootguid']) && array_key_exists($machine['netbootguid'], $action)) {
                    $machine['action'] = $action;
                }
            }
            if (empty($machine['action'])) {
                $machine['action'] = $actions['default'];
            }
        }
    } else {
        // pas de machine AD existante, on crée une action temporaire à mettre en cache avec la mac et l'uuid en attendant d'avoir un nom de machine
        foreach ($actions as $id => $action) {
            if (! empty($uuid) && array_key_exists($uuid, $action)) {
                $machine = array(
                    'networkaddress' => $mac,
                    'action' => $action,
                    'netbootguid' => $uuid,
                    'memberof' => []
                );
            }
        }
        if (empty($machine)) {
            $machine = array(
                'networkaddress' => $mac,
                'action' => [],
                'netbootguid' => $uuid,
                'memberof' => []
            );
        }
    }
    return $machine;
}

/**
 * enregistre l'action pour une machine.
 * - si accès ipxe : token + mac doivent être fournis, eventuellement id, action, role
 * - si accès sysrescuecd : mac suffit, on est déjà enregistré en phase ipxe
 * - si accès interface : choix de tous les parametres.
 *
 * @param array $config
 * @param string $uuid
 *            id unique machine (smbios)
 *            : nom ou ip ou mac de la machine
 * @param string $type
 *            : type d'action ipxe ( install | clonage | clonage2 | wpkg | ltsp | renomme | default )
 * @param string $id
 *            : identifiant de l'action à effectuer (généré automatiquement) commun à toutes les machines de l'action en cours
 * @param string $role
 *            : role de la machine le cas échéant (modele|clone|nouveau nom|windows|linux)
 * @param string $script
 *            : script à exécuter par ipxe (sysrecuecd|sysprep| post | serv_ltsp|deb-*|windows|default). Dans le cas du clonage, contient les noms des étapes intermédiaires.
 * @param int $ret
 *            : etape pour les scripts en plusieurs étapes côté client.
 *            
 * @return array $actions
 */
// function set_action(array $config, string $uuid, string $type = "", string $id = "", string $role = "", string $script = "", int $ret = - 255, array $disks = array(), bool $increment = false)
function set_action(array $config, string $uuid, array $params = [], bool $increment = false)
{
    // on recupere les valeurs
    $id = $params['id'] ?? "";
    lock('actions');
    $machine = get_action($config, $uuid);
    if (! isset($machine['netbootguid'])) {
        unlock('actions');
        return false;
    }
    $uuid = $machine['netbootguid'];
    if (isset($params['type']) && $params['type'] == "default") { // || $script == "default") {
        delete_action_members($uuid);
        $id = "";
    } else {

        // initialisation de l'id
        if (empty($id)) {
            if (! empty($machine['netbootinitialization'])) {
                $id = $machine['netbootinitialization'];
            } elseif (isset($machine['cn'])) {
                $id = $params['type'] . "-" . $machine['cn'] . "@" . date_et_heure();
                // $machine['netbootinitialization'] = $id;
            } else {
                $id = $params['type'] . "-" . $machine['cn'] . "@" . date_et_heure();
            }
        } elseif ($id == "new") {
            $id = $params['type'] . "-" . $machine['cn'] . "@" . date_et_heure();
        }
    }
    $actions = fetch_action();
    $res = false;
    // modifs ad si besoin
    if (! empty($id) && count($machine) > 0 && isset($machine['cn'])) {
        $attrs = array();
        $res = true;
        if (! empty($machine['netbootinitialization'])) {
            if ($id != $machine['netbootinitialization']) {

                // on enleve la machine de l'action
                unset($actions[$machine['netbootinitialization']][$machine['netbootguid']]);
                $attrs['netbootinitialization'] = $id;
                $res &= modify_ad($config, $machine['cn'], "computer", $attrs, "replace");
            }
        } else {
            $attrs['netbootinitialization'] = $id;
            $res &= modify_ad($config, $machine['cn'], "computer", $attrs, "add");
        }
        /*
         * $attrs = array();
         * if (! empty($machine['netbootguid'])) {
         * if ($machine['netbootguid'] != $uuid) {
         *
         * $attrs['netbootguid'] = $uuid;
         * foreach ($actions as $key => $action) {
         * if (isset($action['netbootguid'])) {
         * unset($actions[$key]['netbootguid']);
         * }
         * }
         * $res &= modify_ad($config, $machine['cn'], "computer", $attrs, "replace");
         * }
         * } else {
         * $res = false;
         * $attrs['netbootguid'] = "";
         * // $res &= modify_ad($config, $machine['cn'], "computer", $attrs, "add");
         * }
         */
        if (! $res) {
            file_put_contents("/tmp/ldap_err.log", "machine : " . print_r($machine, true) . " attrs : " . print_r($attrs, true) . "\n", FILE_APPEND);
        } else {
            search_machine($config, $machine['cn'], true, true);
        }
    }
    // mise a jour de actions
    if (! empty($id)) {
        $actions[$id] = $actions[$id] ?? array();
        if (! empty($params['type'])) {
            $actions[$id]['type'] = $params['type'];
        }
        $actions[$id][$uuid] = $actions[$id][$uuid] ?? array();
        if (! empty($params['disks']) && count($params['disks']) > 0) {
            $actions[$id]['disks'] = $params['disks'];
        }
        if (! empty($params['role'])) {
            $actions[$id][$uuid]['role'] = $params['role'];
        }
        if (! empty($params['script'])) {
            $actions[$id][$uuid]['script'] = $params['script'];
        }
        if (! empty($params['etape'])) {
            $actions[$id][$uuid]['etape'] = $params['etape'];
        }
        $actions[$id][$uuid]['ret'] = $params['ret'] ?? $actions[$id][$uuid]['ret'] ?? - 1;
        if ($increment) {
            $actions[$id][$uuid]['ret'] ++;
        }
    }
    apcu_store('actions', $actions);
    unlock('actions');

    return get_action($config, $uuid);
}

function increment_action(array $config, string $cn)
{
    return set_action($config, $cn, [], true);
}

function delete_action_members(string $uuid)
{
    $actions = fetch_action();
    foreach ($actions as $key => $action) {
        if (isset($action[$uuid])) {
            unset($actions[$key][$uuid]);
        }
    }
    return $actions;
}

function delete_action(string $id)
{
    lock('actions');
    $actions = fetch_action();
    if (isset($actions[$id])) {
        unset($actions[$id]);
    }
    apcu_store('actions', $actions);
    unlock('actions');
    return $actions;
}

/**
 * verifie que l'action est authentifiée pour la machine
 *
 * @param array $config
 * @param string $cn
 *            :nom ou ip ou mac de la machine (ou n'importe quel attribut unique)
 * @param string $uuid
 *            : jeton d'identification
 * @return boolean
 */
function auth_action(array $config, string $cn, string $uuid)
{
    $machine = get_action($config, $uuid, $cn);
    return (! empty($machine['netbootguid']) && $machine['netbootguid'] == $uuid);
}

/**
 * retourne true pour le boot iPXE en cas d'autentification avec un compte computer_is_admin et que la machine existe en cache actions
 *
 * @param array $config
 * @param string $login
 * @param string $password
 * @return string $uuid : jeton unique, "" si non autorisé
 */
function login_action(array $config, $uuid, string $login, string $password)
{
    $res = false;
    if (user_valid_passwd($config, $login, $password) && have_right($config, "computer_is_admin", $login)) {
        $machine = get_action($config, $uuid);
        if (! empty($machine['netbootguid'])) {
            $res = true;
        }
        // else {
        // $uuid = bin2hex(random_bytes(32));
        // if (isset($machine['cn'])) {
        // $attrs = array(
        // 'netbootguid' => $uuid
        // );
        // modify_ad($config, $machine['cn'], "computer", $attrs);
        // } else {
        // set_action($config, $cn, "enregistrement", $cn, $uuid);
        // }
        // }
    }
    return $res;
}
?>
