<?php

/**
 * Fonctions pour l'automatisation NextCloud
 * @Projet SambaEdu
 * @Auteurs Equipe Sambaedu
 * @Licence Distribue sous la licence GPL
 * @Repertoire: includes/
 * file: cloud.inc.php
 *
 * Accès minimal à l'API NC pour permettre la création de partages (classes, groupes) et la récupération de la clé client pour le montage RClone
 * 
 * Utilisation de lib php webdav
 * 
 * Seuls les appels utiles sont implémentés.
 * 
 * Authentification
 * 
 * - login avec un compte et sa clé d'api
 * - récupération de la clé api pour un compte utilisateur en passant le mot de passe, ou via un compte admin si c'est possible
 * 
 * Dossiers (api webdav)
 * 
 * - liste des dossiers
 * - mkdir
 * - rm
 * - mv
 * 
 * Partages (api shares)
 * 
 * - liste de droits
 * - ajout d'un droit
 * - suppression d'un droit
 * 
 * 
 * les données de conf sont dans /etc/sambaedu/sambaedu.conf.d/cloud.conf
 * 
 * cloud_name = "nuage"
 * cloud_type = "webdav"
 * cloud_uri = "https://nuage.etab.fr/"
 *
 * si les utilisateurs sont identifiés par leur GUID (et pas leur CN) dans NC
 *
 * cloud_userid = "guid" 
 *
 * un compte admin sur le NC doit être défini dans la conf pour les opérations.
 *
 * cloud_api_user = "nomdeladmincloud"
 * cloud_api_password = "token"
 *
 *
 * la lib sabre/dav est utilisée pour faire les requêtes webdav
 *
 * les partages sont gérés par un utilisateur unique par établissement.
 */
function cloud_path($config, $user)
{
    switch (@$config['cloud_type']) {
        case "seafile":
            $path = $config['cloud_uri'];
            break;
        default:
            $path = $config['cloud_uri'] . "remote.php/dav/files/" . $user . "/";
    }
    return $path;
}

function get_nc_client_code($config, $user, $password)
{
    $info = false;
    require_once (dirname(__FILE__) . '/../vendor/autoload.php');
    $opt = [
        'base_uri' => $config['cloud_uri'],
        'cookies' => true,
        'headers' => [
            'OCS-APIRequest' => "true"
        ]
    ];
    curl_proxy_options($config, $opt);
    $client = new GuzzleHttp\Client($opt);
    try {
        $response = $client->request('GET', 'ocs/v2.php/core/getapppassword', [
            'auth' => [
                $user,
                $password
            ],
            'query' => [
                'format' => 'json'
            ]
        ]);
        if ($response) {
            $res = json_decode((string) $response->getBody(), true);
            if ($res['ocs']['meta']['status'] == "ok") {
                $info['code'] = $res['ocs']['data']['apppassword'];
                $response = $client->request('GET', 'ocs/v2.php/cloud/user', [
                    'auth' => [
                        $user,
                        $info['code']
                    ],
                    'query' => [
                        'format' => 'json'
                    ]
                ]);
                if ($response) {
                    $res = json_decode((string) $response->getBody(), true);
                    if ($res['ocs']['meta']['status'] == "ok") {
                        $info['id'] = $res['ocs']['data']['id'];
                        $info['storageLocation'] = $res['ocs']['data']['storageLocation'];
                    }
                }
            }
        }
    } catch (\GuzzleHttp\Exception\RequestException $e) {
        return false;
    }
    return $info;
}

function test_cloud_login($config, $user = "", $password = "")
{
    if (empty($user) && empty($password)) {
        $creds = [
            $config['cloud_api_user'],
            $config['cloud_api_password']
        ];
    } else {
        $creds = [
            $user,
            $password
        ];
    }

    require_once (dirname(__FILE__) . '/../vendor/autoload.php');
    $opt = [
        'base_uri' => $config['cloud_uri'],
        'cookies' => true,
        'headers' => [
            'OCS-APIRequest' => "true"
        ],
        'query' => [
            'format' => 'json'
        ]
    ];
    curl_proxy_options($config, $opt);
    $client = new GuzzleHttp\Client($opt);
    try {
        $response = $client->request('GET', 'ocs/v2.php/cloud/user?format=json', [
            'auth' => $creds
        ]);
        if ($response) {
            $res = json_decode((string) $response->getBody(), true);
            if ($res['ocs']['meta']['status'] == "ok") {
                $data = $res['ocs']['data'];
            }
        }
    } catch (\GuzzleHttp\Exception\RequestException $e) {
        return $e;
    }
    return $data;
}

function configure_apps_client($config, $user, &$cloud = [])
{
    $data = unserialize($_POST['data'] ?? NULL);
    $nuage = $_POST['nuage'] ?? "";
    $init = $_POST['init'] ?? "0";

    $conf = get_rclone_config($config, $user);
    if (isset($conf['nuage.apps']) && $init == 0) {
        $config['cloud_uri'] = parse_url($conf['nuage.apps']['url'], PHP_URL_SCHEME) . "://" . parse_url($conf['nuage.apps']['url'], PHP_URL_HOST) . "/";
        // $config['cloud_uri'] = "https://nuage.apps.education.fr/";
        $config['cloud_type'] = $conf['nuage.apps']['type'];
        $config['cloud_vendor'] = $conf['nuage.apps']['vendor'];
        $config['cloud_name'] = 'nuage.apps';
        $config['cloud_api_password'] = $conf['nuage.apps']['pass'];
        $config['cloud_api_user'] = $conf['nuage.apps']['user'];
        $html = 'Vous êtes connecté sur <a href="' . $config['cloud_uri'] . '">' . $config['cloud_uri'] . '</a>  avec le compte
<b>' . $config['cloud_api_user'] . '</b>.<br>
Vous trouverez sur votre bureau un dossier <b>nuages\\nuage.apps</b> vous donnant accès à vos données.<p>';
        $html .= 'Vous pouvez réinitialiser cette connexion en cas de problème<br>
Pour plus d&lsquo;explications sur la connexion de votre ordinateur personnel à ce cloud : <a href="https://magistere.education.fr/f709">Formation à l&lsquo;usage du cloud</a>
<form name="cloud" action="" method="post">
<input type=hidden name=nuage value="">
<input type=hidden name=init value="1">
<input type=submit name="valider" value="Réinitialiser">
</form>';
        $cloud[] = "nuage.apps";
    } else {
        $config['cloud_uri'] = "https://" . $nuage . ".apps.education.fr/";
        // $config['cloud_uri'] = "https://nuage.apps.education.fr/";
        $config['cloud_type'] = "webdav";
        $config['cloud_vendor'] = "nextcloud";
        $config['cloud_name'] = "nuage.apps";

        require_once (dirname(__FILE__) . '/../vendor/autoload.php');

        $opt = [
            'base_uri' => $config['cloud_uri'],
            'cookies' => true,
            'headers' => [
                'OCS-APIRequest' => "true",
                'User-Agent' => "SambaEdu"
            ],
            'http_errors' => false
        ];
        curl_proxy_options($config, $opt);
        $html = '<form name="cloud" action="" method="post">';
        if (! $data) {
            if (empty($nuage)) {
                // étape 1 : choix du serveur
                $html .= "Pour activer la connexion automatique avec votre espace personnel sur apps.education.fr,<br>
Vous devez au préalable avoir créé votre compte sur apps.education.fr, et activé l&lsquo;application \"nuage\".<br>
Choisissez le serveur auquel vous êtes connectés :<br>
<select name=nuage onchange=\"this.form.submit()\">
    <option value=''>Choisir le votre serveur</option>
    <option value='nuage01'>nuage01</option>
    <option value='nuage02'>nuage02</option>
    <option value='nuage03'>nuage03</option>
</select>";
            } elseif (empty($data['poll']['token'])) {
                // Etape 2 : récup de l'url de la page de login et du token associé
                $client = new GuzzleHttp\Client($opt);
                try {
                    $response = $client->request('POST', 'index.php/login/v2', []);
                    if ($response) {
                        // on ouvre la page de login
                        $data = json_decode((string) $response->getBody(), true);
                        $url = $data['login'];
                        $html .= "<input type=hidden name=data value=" . serialize($data) . ">
Pour activer la connexion automatique avec votre espace personnel sur apps.education.fr, Entrez vos identifiants dans la fenêtre qui va apparaître.<br>
Vous devez au préalable avoir créé votre compte sur apps.education.fr, et activé l&lsquo;application \"nuage\".<br>
Le mot de passe saisi sert juste à la vérification et ne sera pas enregistré<p>
<script type=\"text/javascript\">
    function openOnce(url, target){
      var winref = window.open('', target, 'resizable=yes,height=500,width=500');
      if(winref.location.href === 'about:blank'){
         winref.location.href = url;
         var refresh = true;
      }else if(winref.location.href === '" . $url . "/#/'){
         var refresh = true;
      }else{
         var refresh = false;
      }
      return winref;
    }
    openOnce('" . $url . "', 'apps.education');
    var wait = setTimeout(document.cloud.submit(), 29000);
</script>
<input type=hidden name=data value=" . serialize($data) . ">
<input type=hidden name=nuage value=" . $nuage . ">
<input type=\"submit\" name=\"valider\" value=\"Valider\">";
                    } else {
                        $html .= "erreur, pas de réponse !";
                    }
                } catch (\GuzzleHttp\Exception\RequestException $e) {
                    $html .= "erreur : " . print_r($e, true);
                }
            }
        } else {
            $client = new GuzzleHttp\Client($opt);
            try {
                $response = $client->request('POST', $data['poll']['endpoint'], [
                    'form_params' => [
                        'token' => $data['poll']['token']
                    ]
                ]);
                if ($response->getStatusCode() == 200) {
                    $data = json_decode((string) $response->getBody(), true);
                    $config['cloud_uri'] = $data['server'] . "/";
                    $info = [
                        'id' => $data['loginName'],
                        'code' => $data['appPassword']
                    ];
                    $res = create_rclone_config($config, $info, $user);
                    if ($res === true) {
                        $html .= "<p>Votre connexion à <a href='https://apps.education.fr'>apps.education.fr</a> a bien été activée.<br>
Vous trouverez sur votre bureau un dossier <b>nuages\\nuage.apps</b> vous donnant accès à vos données.<p>";
                    } else {
                        $html .= $res;
                    }
                } elseif ($response->getStatusCode() == 404) {
                    $html .= "Attente de l'authentification sur nuage<br> 
<input type=hidden name=data value=" . serialize($data) . ">
<script type=\"text/javascript\">
    var wait = setTimeout(document.cloud.submit(), 29000)
</script>
<input type=hidden name=nuage value=" . $nuage . ">
<input type=\"submit\" name=\"valider\" value=\"Valider\">";
                } else {
                    $html .= "erreur : " . $response->getStatusCode();
                }
            } catch (\GuzzleHttp\Exception\ConnectException $e) {
                $html .= "erreur : <pre>" . print_r($e, true) . "</pre>";
            }
        }
        $html .= '</form>';
    }
    return $html;
}

function get_seafile_client_code($config, $user, $password)
{
    $code = false;
    require_once (dirname(__FILE__) . '/../vendor/autoload.php');
    $opt = [
        'base_uri' => $config['cloud_uri'],
        'cookies' => true
    ];
    curl_proxy_options($config, $opt);
    $client = new GuzzleHttp\Client($opt);
    try {
        $response = $client->request('POST', 'api2/auth-token/', [
            'form_params' => [
                'username' => $user . "@" . $config['domain'],
                'password' => $password
            ]
        ]);
        if ($response) {
            $res = json_decode((string) $response->getBody(), true);
            if ($res['token']) {
                $info['code'] = $res['token'];
                $info['id'] = $user;
            }
        }
    } catch (\GuzzleHttp\Exception\RequestException $e) {
        return false;
    }
    return $info;
}

function test_rclone_config($config, $login)
{
    $rclone_config = "/home/$login/.config/rclone/rclone.conf";
    exec("sudo stat " . $rclone_config . " 2>&1", $out, $ret);
    if ($ret == 0) {
        $rclone = "rclone lsjson " . $config['cloud_name'] . ":/ --max-depth=1 --config=$rclone_config 2>/dev/null";
        exec("sudo " . $rclone, $out, $ret);
        if ($ret == 0) {
            return true;
        } else {
            return false;
        }
    } else {
        return false;
    }
}

function create_rclone_config($config, $info, $login = "")
{
    if (empty($login))
        $login = $info['id'];
    $rclone_config = "/home/$login/.config/rclone/rclone.conf";
    system("sudo mkdir -p /home/$login/.config/rclone/");
    system("sudo chown -R $login /home/$login/.config/");

    switch (@$config['cloud_type']) {
        case "seafile":
            $rclone = "rclone config create " . $config['cloud_name'] . " " . $config['cloud_type'] . " vendor=sambaedu url=\"" . cloud_path($config, $info['id']) . "\" token=" . $info['code'] . " --obscure --config=$rclone_config";
            break;
        default:
            $rclone = "rclone config create " . $config['cloud_name'] . " " . $config['cloud_type'] . " vendor=sambaedu url=\"" . cloud_path($config, urlencode($info['id'])) . "\" user=\"" . $info['id'] . "\" pass=" . $info['code'] . " --obscure --config=$rclone_config";
    }
    exec("sudo " . $rclone, $out, $ret);
    if ($ret == 0) {
        exec("sudo chown $login $rclone_config");
        groupaddmember($config, $login, "Cloud");
        $ret = test_rclone_config($config, $login);
    } else {
        $ret = false;
    }
    // return "erreur creation rclone : <pre> $rclone " . implode("<br>", $out) . "</pre>";
    return $ret;
}

function configure_user_cloud($config, $user, $password)
{
    $ret = true;
    switch (@$config['cloud_type']) {
        case "seafile":
            // $code = $password;
            $info = get_seafile_client_code($config, $user, $password);
            break;
        case "webdav":
            $info = get_nc_client_code($config, $user, $password);
            $info['id'] = $user;
            $info['storageLocation'] = cloud_path($config, $user);
            break;
        default:
            $info = false;
    }
    if ($info) {
        $ret = create_rclone_config($config, $info);
    }
    return $ret;
}

function get_nc_userdata($config, $user)
{
    require_once (dirname(__FILE__) . '/../vendor/autoload.php');
    $opt = [
        'base_uri' => $config['cloud_uri'],
        'cookies' => true,
        'headers' => [
            'OCS-APIRequest' => "true"
        ]
    ];
    curl_proxy_options($config, $opt);
    $client = new GuzzleHttp\Client($opt);
    try {
        $response = $client->request('GET', 'ocs/v2.php/cloud/users/' . $user, [
            'auth' => [
                $config['cloud_api_user'],
                $config['cloud_api_password']
            ],
            'query' => [
                'format' => 'json'
            ]
        ]);
        if ($response) {
            $res = json_decode((string) $response->getBody(), true);
            if ($res['ocs']['meta']['status'] == "ok") {
                $data = $res['ocs']['data'];
            }
        }
    } catch (\GuzzleHttp\Exception\RequestException $e) {
        return false;
    }
    return $data;
}

function search_nc_user($config, $user)
{
    if (is_array($user)) {
        $u = $user['cn'];
    } else {
        $u = $user;
    }

    if (! get_nc_userdata($config, $u)) {

        if (! is_array($user)) {
            $ad = search_user($config, $user);
        } else {
            $ad = $user;
        }
        if (! $ad) {
            return false;
        }
        $fullname = $ad['fullname'] ?? $ad['nom'] ?? explode('@', $ad['mail'])[0] ?? explode('@', $ad['userprincipalname'])[0];
        require_once (dirname(__FILE__) . '/../vendor/autoload.php');
        $opt = [
            'base_uri' => $config['cloud_uri'],
            'cookies' => true,
            'headers' => [
                'OCS-APIRequest' => "true"
            ]
        ];
        curl_proxy_options($config, $opt);
        $client = new GuzzleHttp\Client($opt);
        try {
            $response = $client->request('GET', 'ocs/v2.php/core/autocomplete/get', [
                'auth' => [
                    $config['cloud_api_user'],
                    $config['cloud_api_password']
                ],
                'query' => [
                    'format' => 'json',
                    'search' => $fullname,
                    'limit' => 1
                ]
            ]);
            if ($response) {
                $res = json_decode((string) $response->getBody(), true);
                if ($res['ocs']['meta']['status'] == "ok") {
                    $u = $res['ocs']['data'][0]['id'] ?? false;
                }
            }
        } catch (\GuzzleHttp\Exception\RequestException $e) {
            return false;
        }
    }
    return $u;
}

function get_rclone_config($config, $user)
{
    $rclone_config = "/home/$user/.config/rclone/rclone.conf";
    $rclone = "rclone config dump --config=$rclone_config";
    exec("sudo " . $rclone, $out, $ret);
    if ($ret == 0) {
        return json_decode(implode(' ', $out), true);
    }
    return false;
}

define('OCS_GROUP', 1);
define('OCS_USER', 0);
define('OCS_PERM_RO', 1);
define('OCS_PERM_RW', 15);

function get_nc_all_shares($config)
{
    require_once (dirname(__FILE__) . '/../vendor/autoload.php');
    $opt = [
        'base_uri' => $config['cloud_uri'],
        'cookies' => true,
        'headers' => [
            'OCS-APIRequest' => "true"
        ]
    ];
    curl_proxy_options($config, $opt);
    $client = new GuzzleHttp\Client($opt);
    try {
        $response = $client->request('GET', 'ocs/v2.php/apps/files_sharing/api/v1/shares', [
            'auth' => [
                $config['cloud_api_user'],
                $config['cloud_api_password']
            ],
            'query' => [
                'format' => 'json'
            ]
        ]);
        if ($response) {
            $res = json_decode((string) $response->getBody(), true);
            if ($res['ocs']['meta']['status'] == "ok") {
                $data = $res['ocs']['data'];
            }
        }
    } catch (\GuzzleHttp\Exception\RequestException $e) {
        return $e;
    }
    return $data;
}

function get_nc_shares($config, $path, $sub = 'false', &$client = NULL)
{
    require_once (dirname(__FILE__) . '/../vendor/autoload.php');
    if (! $client) {
        $opt = [
            'base_uri' => $config['cloud_uri'],
            'cookies' => true,
            'headers' => [
                'OCS-APIRequest' => "true"
            ]
        ];
        curl_proxy_options($config, $opt);
        $client = new GuzzleHttp\Client($opt);
    }
    try {
        $response = $client->request('GET', 'ocs/v2.php/apps/files_sharing/api/v1/shares', [
            'auth' => [
                $config['cloud_api_user'],
                $config['cloud_api_password']
            ],
            'query' => [
                'path' => $path,
                'format' => 'json',
                'subfiles' => $sub
            ]
        ]);
        if ($response) {
            $res = json_decode((string) $response->getBody(), true);
            if ($res['ocs']['meta']['status'] == "ok") {
                $data = $res['ocs']['data'];
            }
        }
    } catch (\GuzzleHttp\Exception\RequestException $e) {
        return false;
    }
    return $data;
}

function create_nc_share($config, $path, $share, &$data = [])
{
    $u = $share['share_with'];
    if ($share['share_type'] == OCS_GROUP || $u = search_nc_user($config, $u)) {

        require_once (dirname(__FILE__) . '/../vendor/autoload.php');
        $opt = [
            'base_uri' => $config['cloud_uri'],
            'cookies' => true,
            'headers' => [
                'OCS-APIRequest' => "true"
            ]
        ];
        curl_proxy_options($config, $opt);
        $client = new GuzzleHttp\Client($opt);
        try {
            $response = $client->request('POST', 'ocs/v2.php/apps/files_sharing/api/v1/shares', [
                'auth' => [
                    $config['cloud_api_user'],
                    $config['cloud_api_password']
                ],
                'form_params' => [
                    'path' => $path,
                    'shareType' => $share['share_type'],
                    'shareWith' => $u,
                    'permissions' => $share['permissions'],
                    'format' => 'json'
                ]
            ]);
            if ($response) {
                $res = json_decode((string) $response->getBody(), true);
                $data = $res['ocs']['data'];
                if ($res['ocs']['meta']['status'] == "ok") {
                    return true;
                } else {
                    return false;
                }
            }
        } catch (\GuzzleHttp\Exception\RequestException $e) {
            return false;
        }
        return true;
    } else {
        return false;
    }
}

function modify_nc_share($config, $id, $share, &$data = [])
{
    if ($u = search_nc_user($config, $share['share_with'])) {

        require_once (dirname(__FILE__) . '/../vendor/autoload.php');
        $opt = [
            'base_uri' => $config['cloud_uri'],
            'cookies' => true,
            'headers' => [
                'OCS-APIRequest' => "true"
            ]
        ];
        curl_proxy_options($config, $opt);
        $client = new GuzzleHttp\Client($opt);
        try {
            $response = $client->request('PUT', 'ocs/v2.php/apps/files_sharing/api/v1/shares/' . $id, [
                'auth' => [
                    $config['cloud_api_user'],
                    $config['cloud_api_password']
                ],
                'query' => [
                    'permissions' => $share['permissions'],
                    'format' => 'json'
                ]
            ]);
            if ($response) {
                $res = json_decode((string) $response->getBody(), true);
                $data = $res['ocs']['data'];
                if ($res['ocs']['meta']['status'] == "ok") {
                    return true;
                } else {
                    return false;
                }
            }
        } catch (\GuzzleHttp\Exception\RequestException $e) {
            return false;
        }
        return true;
    } else {
        return false;
    }
}

function delete_nc_share($config, $id)
{
    require_once (dirname(__FILE__) . '/../vendor/autoload.php');
    $opt = [
        'base_uri' => $config['cloud_uri'],
        'cookies' => true,
        'headers' => [
            'OCS-APIRequest' => "true"
        ]
    ];
    curl_proxy_options($config, $opt);
    $client = new GuzzleHttp\Client($opt);
    try {
        $response = $client->request('DELETE', 'ocs/v2.php/apps/files_sharing/api/v1/shares/' . $id, [
            'auth' => [
                $config['cloud_api_user'],
                $config['cloud_api_password']
            ],
            'query' => [
                'format' => 'json'
            ]
        ]);
        if ($response) {
            $res = json_decode((string) $response->getBody(), true);
            $data = $res['ocs']['data'];
            if ($res['ocs']['meta']['status'] == "ok") {
                return true;
            } else {
                return false;
            }
        }
    } catch (\GuzzleHttp\Exception\RequestException $e) {
        return false;
    }
    return true;
}

// webdav
function dav_proxy_options($config, &$opt = [])
{
    if (! empty($config['server_proxy']) && ! empty($config['proxy_address']) && ! empty($config['proxy_port'])) {
        $opt['proxy'] = "http://" . $config['proxy_address'] . ":" . $config['proxy_port'];
    }
}

/**
 * liste l'ensemble des dossiers avec leurs partages éventuels
 *
 * fonction destinée à la synchro des dossiers groupes. Pas forcément adaptée à autre chose.
 * retourne un tableau avec les élements compatibles avec la création ou la suppression.
 *
 * @param unknown $config
 * @param string $path
 * @param unknown $client
 * @param number $depth
 * @return \Sabre\HTTP\ClientHttpException|array
 */
function list_nc_files($config, $path = "", &$client = NULL, $depth = 3)
{
    require_once (dirname(__FILE__) . '/../vendor/autoload.php');

    $user = $config['cloud_api_user'];

    if (! $client) {

        $opt = array(
            'baseUri' => cloud_path($config, $user),
            // 'userName' => $user,
            'userName' => $config['cloud_api_user'],
            'password' => $config['cloud_api_password']
        );
        dav_proxy_options($config, $opt);
        $client = new Sabre\DAV\Client($opt);
    }

    try {
        $data = $client->propfind($path, array(
            '{http://owncloud.org/ns}share-types',
            '{http://owncloud.org/ns}owner-display-name',
            '{http://owncloud.org/ns}fileid'
        ), $depth);
        // return $data;
    } catch (Sabre\HTTP\ClientHttpException $e) {
        return $e;
    }
    $shares = get_nc_all_shares($config);
    $list = [];
    foreach ($data as $dav_path => $info) {
        $dav_path = urldecode($dav_path);
        preg_match("!/remote\.php/dav/files/" . $user . "(/?" . $path . ".*)/!", $dav_path, $m);
        $relative_path = $m[1] ?? "";
        if (! empty($relative_path)) {
            $keys = array_keys(array_column($shares, 'path'), $relative_path);
            if (count($keys) > 0) {
                foreach ($keys as $key) {
                    $list[$relative_path][$shares[$key]['id']] = [
                        'type' => $shares[$key]['share_type'],
                        'permissions' => $shares[$key]['permissions'],
                        'share_with' => $shares[$key]['share_with']
                    ];
                }
            } else {
                $list[$relative_path] = [];
            }
        }
    }
    return $list;
}

function create_nc_folder($config, $path, &$client)
{
    require_once (dirname(__FILE__) . '/../vendor/autoload.php');

    if (! $client) {
        $user = $config['cloud_api_user'];

        $opt = array(
            'baseUri' => cloud_path($config, $user),
            'userName' => $user,
            'password' => $config['cloud_api_password']
        );
        dav_proxy_options($config, $opt);
        $client = new Sabre\DAV\Client($opt);
    }

    try {
        $data = $client->request('MKCOL', ltrim($path, "/"));
        $m = "";
    } catch (Sabre\DAV\Exception $e) {
        $m = "Erreur :";
    }
    return $m . $data['statusCode'];
}

function delete_nc_folder($config, $path)
{
    require_once (dirname(__FILE__) . '/../vendor/autoload.php');

    $user = $config['cloud_api_user'];

    $opt = array(
        'baseUri' => cloud_path($config, $user),
        'userName' => $user,
        'password' => $config['cloud_api_password']
    );
    dav_proxy_options($config, $opt);
    $client = new Sabre\DAV\Client($opt);

    try {
        $data = $client->request('DELETE', ltrim($path, "/"));
        if ($data['statusCode'] == '201') {
            return true;
        }
    } catch (Sabre\DAV\Exception $e) {
        return false;
    }
    return true;
}

function move_nc_folder($config, $path, $newpath)
{
    require_once (dirname(__FILE__) . '/../vendor/autoload.php');

    $user = $config['cloud_api_user'];

    $opt = array(
        'baseUri' => cloud_path($config, $user),
        'userName' => $user,
        'password' => $config['cloud_api_password']
    );
    dav_proxy_options($config, $opt);
    $client = new Sabre\DAV\Client($opt);
    try {
        $data = $client->request('MOVE', ltrim($path, "/"), NULL, [
            'Destination' => cloud_path($config, $user) . ltrim($newpath, "/"),
            'Overwrite' => 'T'
        ]);
        if ($data['statusCode'] == '201') {
            return true;
        }
    } catch (Sabre\DAV\Exception $e) {
        return false;
    }
    return true;
}

function list_nc_echanges($config, $prof)
{
    $echanges = [];
    if (is_array($prof)) {
        $classes = $prof['memberof'];
    } else {
        $classes = search_user($config, $prof)['memberof'];
    }
    foreach ($classes as $classe) {
        $m = array();

        if (preg_match("/CN=Equipe_(.*),OU=equipe.*/i", $classe, $m)) {
            $nom = $m[1];
            if (count(search_ad($config, $nom, "classe")) > 0) {
                $type = "Classe";
            } else {
                $type = "Cours";
            }
            $res = get_nc_shares($config, "/Groupes/@" . $nom . "/Echange@" . $nom);
            $echanges[$nom] = [
                'id' => 0,
                'type' => "Classe"
            ];
            if (count($res) > 0) {
                foreach ($res as $s) {
                    if ($s['share_type'] == OCS_GROUP && $s['permissions'] == OCS_PERM_RW && $s['share_with'] == $type . "_" . $nom . $config['suffix']) {
                        $echanges[$nom] = [
                            'id' => $s['id'],
                            'type' => $type
                        ];
                        break;
                    }
                }
            }
        }
    }

    return $echanges;
}

function set_nc_echanges($config, &$echanges)
{
    foreach ($echanges as $nom => $e) {
        if ($e['id'] == 0) {
            $res = delete_nc_share($config, $e['id']);
        } else {
            $s = [
                'share_type' => OCS_GROUP,
                'share_with' => $e['type'] . "_" . $nom . $config['suffix'],
                'permissions' => OCS_PERM_RW
            ];
        }

        $res = create_nc_share($config, "/Groupes/@" . $nom . "/Echange@" . $nom, $s);
        $res &= get_nc_shares($config, "/Groupes/@" . $nom . "/Echange@" . $nom);
        if (count($res) > 0) {
            foreach ($res as $s) {
                if ($s['share_type'] == OCS_GROUP && $s['permissions'] == OCS_PERM_RW && $s['share_with'] == $e['type'] . "_" . $nom . $config['suffix']) {
                    $echanges[$nom] = [
                        'id' => $s['id'],
                        'type' => $e['type']
                    ];
                    break;
                }
            }
        }
    }
    return $res;
}

/**
 * Bascule un partage échange entre l'état partage et non partagé
 *
 * @param unknown $config
 * @param unknown $nom
 * @param unknown $echanges
 *            liste des Echanges en cache
 * @return boolean
 */
function toggle_nc_echange($config, $nom, &$echanges)
{
    $e = $echanges[$nom];
    if ($e['id'] != 0) {
        $res = delete_nc_share($config, $e['id']);
        $echanges[$nom] = [
            'id' => 0,
            'type' => $e['type']
        ];
    } else {
        $s = [
            'share_type' => OCS_GROUP,
            'share_with' => $e['type'] . "_" . $nom . $config['suffix'],
            'permissions' => OCS_PERM_RW
        ];
        if ($res = create_nc_share($config, "/Groupes/@" . $nom . "/Echange@" . $nom, $s)) {
            $res = get_nc_shares($config, "/Groupes/@" . $nom . "/Echange@" . $nom);
            if (count($res) > 0) {
                foreach ($res as $s) {
                    if ($s['share_type'] == OCS_GROUP && $s['permissions'] == OCS_PERM_RW && $s['share_with'] == $e['type'] . "_" . $nom . $config['suffix']) {
                        $echanges[$nom] = [
                            'id' => $s['id'],
                            'type' => $e['type']
                        ];
                        break;
                    }
                }
            }
        }
    }
    return $res;
}

function path_sort($a, $b){
    return count(explode('/', $a)) - count(explode('/', $b));
}

/**
 * calcule le tableau des différences entre deux états des partages a et b
 * génère un tableau de la forme ['path' => ['id'=> [share]]]
 * si 'path'=> [], on déplace ou efface le dossier
 * si 'id' = 0, on crée le partage
 * si 'id' != 0 :
 * si share = [] on supprime le partage
 * sinon on le modifie
 *
 * Ne modifie pas le partage des dossiers Echange, mais les crée ou supprime si besoin.
 *
 * attention l'algo ne fonctionne que si il n'y a qu'un seul partage par dossier !!!
 *
 * @param unknown $aa
 *            : état souhaité
 * @param unknown $bb
 *            : état à soustraire
 * @param boolean $del
 *            : si true on prend en compte les dossiers dans archives pour la suppression
 * @return unknown[]
 */
function diff_shares($config, $aa, $bb, $del = false)
{
    $diff = [];
    foreach ($aa as $pa => $a) {
        $m = false;
         // pas de prise en compte des dossiers déjà dans Archives
        if (! preg_match("|/Archives/|", $pa) || $del) {
            foreach ($bb as $pb => $b) {
                if ($pa == $pb) {
                    if (preg_match("|/Echange@|", $pa)) {
                        $m = true;
                        break;
                    } else {
                        if (count($a) > 0) {
                            foreach ($a as $ida => $sa) {
                                if (count($b) > 0) {
                                    foreach ($b as $idb => $sb) {
                                        if ($ida == $idb) {
                                            // même share id, on modifie
                                            $s = array_diff($sa, $sb);
                                            if (count($s) > 0) {
                                                $diff[$pa][$ida] = $sa;
                                            }
                                        } elseif ($ida == 0) {
                                            if ($sa['share_with'] != $sb['share_with']) {
                                                if ($u = search_nc_user($config, $sa['share_with'])) {
                                                    if ($u != $sb['share_with']) {
                                                        $sa['share_with'] = $u;
                                                        // plus de partage, ou changemnt de proprio on efface $idb
                                                        $diff[$pa][$idb] = [];
                                                        // on cree $ida
                                                        $diff[$pa][$ida] = $sa;
                                                    }
                                                }
                                            } else {
                                                // on modifie $idb
                                                $s = array_diff($sa, $sb);
                                                if (count($s) > 0) {
                                                    $diff[$pa][$idb] = $sa;
                                                }
                                            }
                                        }
                                    }
                                } else {
                                    // on cree $ida
                                    $diff[$pa][$ida] = $sa;
                                }
                            }
                        }
                        $m = true;
                        break;
                    }
                }
            }
            if (! $m)
                $diff[$pa] = $a;
        }
    }
    uksort($diff, 'path_sort');
    return $diff;
}

/**
 * mise à jour de l'arborescence du partage pour un groupe.
 * crée les dossiers manquants :
 * Groupes/@groupe -> partagé avec Equipe_groupe
 * Groupes/@groupe/eleve@groupe -> partagé avec eleve
 * déplace les dossiers en trop dans Archive/eleve@groupe@archive -> partagé avec eleve
 *
 * La mise a jour de l'arborescence côté utilisateur est faite par un script de logon sur le montage rclone
 *
 * fonction agissant sur un cache
 */
function update_nc_partages($config, &$html, $delete)
{
    $end = false;
    $time = time();
    $partages = apcu_fetch('partages');
    if (! $partages) {
        $liste = list_nc_files($config, "", $client, 3);
        $groupes = list_partages_groups($config);
        $add = diff_shares($config, $groupes, $liste);
        $remove = diff_shares($config, $liste, $groupes, $delete);
        $html .= count($remove) . " répertoires à supprimer<br>";
        $html .= count($add) . " répertoires à créer<br>";
        $html .= "<pre>" . print_r($remove, true) . "</pre>\n";
        $html .= "<pre>" . print_r($add, true) . "</pre>\n";
    } else {
        $add = $partages['add'];
        $remove = $partages['remove'];
        $html .= count($remove) . " répertoires restants à supprimer<br>";
        $html .= count($add) . " répertoires restants à créer<br>";
    }
    while (count($remove) > 0 && time() < $time + 25) {
        $p = key($remove);
        array_shift($remove);
        $m = [];
        preg_match("!/Groupes/@(.*)/?(.*)@?(.*)!", $p, $m);
        if (! empty($m[2])) {
            if ($m[2] == "Prof") {
                $html .= "deplacement de $p vers  /Groupes/@" . $m[1] . "/Equipe@" . $m[3] . "<br>\n";
                $res = move_nc_folder($config, $p, "/Groupes/@" . $m[1] . "/Equipe@" . $m[3]);
            } else {
                $html .= "deplacement de $p vers  /Archives/" . $m[2] . "@" . $m[3] . "@Archives<br>\n";
                $res = move_nc_folder($config, $p, "/Archives/" . $m[2] . "@" . $m[3] . "@Archives");
            }
            if ($res == "201") {
                $html .= "OK<br>\n";
            } else {
                $html .= " ERREUR  $res<br>\n";
            }
        } elseif (! empty($m[1])) {
            $html .= "effacement de $p <br>\n";
            $res = delete_nc_folder($config, $p);
            if ($res == "201") {
                $html .= "OK<br>\n";
            } else {
                $html .= " ERREUR  $res<br>\n";
            }
        }
        if ($delete) {
            $m = [];
            preg_match("!/Archives/(.*)!", $p, $m);
            if (! empty($m[1])) {
                $html .= "effacement de $p <br>\n";
                $res = delete_nc_folder($config, $p);
                if ($res == "201") {
                    $html .= "OK<br>\n";
                } else {
                    $html .= " ERREUR  $res<br>\n";
                }
            }
        }
    }
    $client = NULL;
    while (count($add) > 0 && time() < $time + 25) {
        $p = array_key_first($add);
        $share = array_shift($add);
        $res1 = create_nc_folder($config, $p, $client);
        if ($res1 == "201") {
            $msg = "creation de $p, \n";
        } else {
            $msg = "creation de $p ERREUR  $res1, \n";
            $res1 = false;
        }
        if (is_array($share)) {
            $res = true;
            foreach ($share as $id => $s) {
                if ($id != 0) {
                    if (count($s) == 0) {
                        $res &= delete_nc_share($config, $id);
                    } else {
                        $res &= modify_nc_share($config, $id, $s, $data);
                    }
                } else {
                    $res &= create_nc_share($config, $p, $s, $data);
                }
                // $html .= "<pre>" . print_r($data, true) . "</pre>\n";
            }
        }
        if ($res) {
            $html .= "partage de $p OK<br>\n";
        } else {
            if (! $res1) {
                $html .= $msg;
            }
            $html .= "partage de $p ERREUR<br>\n";
            // $client2 = NULL;
            // $html .= "<pre>" . print_r(get_nc_shares($config, $p, $client2), true) . "</pre>\n";
        }
    }
    $partages = [
        'add' => $add,
        'remove' => $remove
    ];
    apcu_store('partages', $partages, 1000);
    if ((count($add) + count($remove)) == 0) {
        $end = true;
    }
    if ($end) {
        apcu_delete('partages');
    }
    return $end;
}

/**
 * génération du script qui va déplacer les dossiers partagés de l'élève dans son arborescence.
 * Lancé par les scripts applications, une fois le partage monté par rclone.
 * Nextcloud place les dossiers partagés à la racine, on les déplace pour que cela soit compréhensible.
 * Pour éviter les collisions avec des fichiers de l'utilisateur, ils sont nommés :
 * Groupes/@groupe/Travail@groupe
 * /x@groupe
 * /...@groupe
 * En cas de suppression des dossiers par l'utilisateur, ils sont remis en place au logon suivant (ce ne sont que des liens vers les données réelles du NC)
 *
 * l'élève verra par exemple :
 * Mes travaux\1G1
 * \PhysSpe
 * Travail demandé\1G1
 * \PhysSpe
 * Echanges\1G1
 * \PhysSpe
 *
 * le prof verra par exemple :
 * Mes groupes\1G1\...
 * \TSTI1\Profs@TSTI1
 * \Travail@TSTI1
 * \Echange@TSTI1
 * \x@TSTI1
 * \y@TSTI1
 * \NSI\...
 * \I2D\...
 *
 * Il n'est pas possible pour les profs de renommer les sous dossiers individuels car cela poserait des conflits pour les élèves. seuls les liens de partage peuvent être renommés.
 *
 * Les droits et la visibilité sont définis par l'API, et donc si le dossier échange par exemple est non partagé il ne sera pas vu et créé.
 *
 * @param unknown $config
 * @param unknown $id
 * @return string
 */
function genere_script_partage_utilisateur($config, $id)
{
    if ($info = apcu_fetch("apps." . $id)) {
        if (in_array('Profs', $info['list_u'])) {
            if ($info['os'] == "windows") {
                $path = "%userprofile%\\nuages\\" . $config['cloud_name'] . "\\Mes groupes\\";
                $script = "Set _seconds=0\r
:waitloop\r
  Set /a \"_seconds=_seconds+5\">nul\r
  PING -n 6 127.0.0.1>nul\r
  if %_seconds%==30 goto nextstep\r
  REM if not exist \"" . $path . "\" goto waitloop\r
:nextstep\r
";
                $script .= "if not exist \"" . $path . "\" (md \"" . $path . "\")\r\n";
            } else {
                $path = '/home/\$USER/nuages/' . $config['cloud_name'] . '/';
            }

            foreach ($info['list_u'] as $groupe) {
                if (preg_match("/Equipe_(.*)/", $groupe, $m)) {
                    if ($info['os'] == "windows") {
                        $script .= "if exist \"%userprofile%\\nuages\\" . $config['cloud_name'] . "\\@" . $m[1] . "\" ( if not exist \"" . $path . $m[1] . "\" (move \"%userprofile%\\nuages\\" . $config['cloud_name'] . "\\@" . $m[1] . "\" \"" . $path . $m[1] . "\"))\r\n";
                    } else {}
                }
            }
            $script .= "\r\n";
            $script .= "powershell.exe -ExecutionPolicy Bypass -NoLogo -NonInteractive -NoProfile -Command \"\$o = new-object -com shell.application; \$o.Namespace('" . $path . "').Self.InvokeVerb('pintohome')\"\r\n";
            return $script;
        } else {
            if ($info['os'] == "windows") {
                $path = '%userprofile%\\nuages\\' . $config['cloud_name'] . '\\';
                $script = "Set _seconds=0\r
:waitloop\r
  Set /a \"_seconds=_seconds+5\">nul\r
  PING -n 6 127.0.0.1>nul\r
  if %_seconds%==30 goto nextstep\r
  REM if not exist \"" . $path . "Mes travaux\" goto waitloop\r
:nextstep\r
";
                $script .= "if exist \"" . $path . "\" (\r\n";
                $script .= "if not exist \"" . $path . "Mes travaux\" (md \"" . $path . "Mes travaux\")\r\n";
                $script .= "if not exist \"" . $path . "Travail demandé\" (md \"" . $path . "Travail demandé\")\r\n";
                $script .= "if not exist \"" . $path . "Echange\" (md \"" . $path . "Echange\")\r\n";
            } else {
                $path = '/home/\$USER/nuages/' . $config['cloud_name'] . '/';
            }

            foreach ($info['list_u'] as $groupe) {
                if (preg_match("/(Classe_|Cours_)(.*)/", $groupe, $m)) {
                    if ($info['os'] == "windows") {
                        $script .= "if exist \"" . $path . $info['user']['cn'] . "@" . $m[2] . "\" (if not exist \"" . $path . "Mes travaux\\" . $m[2] . "\" (move /y \"" . $path . $info['user']['cn'] . "@" . $m[2] . "\" \"" . $path . "Mes travaux\\" . $m[2] . "\"))\r\n";
                        $script .= "if exist \"" . $path . "Travail@" . $m[2] . "\" (if not exist \"" . $path . "Travail demandé\\" . $m[2] . "\" (move  /y \"" . $path . "Travail@" . $m[2] . "\" \"" . $path . "Travail demandé\\" . $m[2] . "\"))\r\n";
                        $script .= "if exist \"" . $path . "Echange@" . $m[2] . "\" (if not exist \"" . $path . "Echange\\" . $m[2] . "\" (move /y \"" . $path . "Echange@" . $m[2] . "\" \"" . $path . "Echange\\" . $m[2] . "\"))\r\n";
                        $script .= "powershell.exe -ExecutionPolicy Bypass -NoLogo -NonInteractive -NoProfile -Command \"\$o = new-object -com shell.application; \$o.Namespace('" . $path . "Mes travaux').Self.InvokeVerb('pintohome')\"\r\n";
                        $script .= "powershell.exe -ExecutionPolicy Bypass -NoLogo -NonInteractive -NoProfile -Command \"\$o = new-object -com shell.application; \$o.Namespace('" . $path . "Travail demandé').Self.InvokeVerb('pintohome')\"\r\n";
                        $script .= "powershell.exe -ExecutionPolicy Bypass -NoLogo -NonInteractive -NoProfile -Command \"\$o = new-object -com shell.application; \$o.Namespace('" . $path . "Echange').Self.InvokeVerb('pintohome')\"\r\n";
                    } else {}
                }
            }
            $script .= ")\r\n";
            if ($info['os'] == "windows") {
                $script = iconv("UTF-8", "CP850", $script);
            }
            return $script;
        }
    }
}

/**
 * construit un tableau avec les utilisateurs classés par catégorie en classe
 * pour l'affichage des listes de selection
 *
 * @param unknown $config
 * @param unknown $users
 * @return array|unknown
 */
function list_partages_groups($config, $user = "*")
{
    $shares = [
        '/Groupes' => [],
        '/Archives' => []
    ];
    $users = search_ad($config, $user, "user", $config['dn']['people']);
    foreach ($users as $user) {
        if (isset($user['memberof'])) {
            if (@$config['cloud_userid'] == "guid") {
                $u = strtoupper(to_guid($user['objectguid']));
            } else {
                $u = $user['cn'];
                // if (! $u = search_nc_user($config, $user)) {
                // continue;
                // }
            }
            $i = 0;
            foreach ($user['memberof'] as $groupe) {
                $m = array();
                if (preg_match("/CN=(Classe|Equipe|Cours)_(.*),OU=(Classes|Equipes|Cours).*/i", $groupe, $m)) {
                    if ($i == 0) {
                        $shares['/Groupes/@' . $m[2]] = [
                            [
                                'share_type' => OCS_GROUP,
                                'permissions' => OCS_PERM_RW,
                                'share_with' => "Equipe_" . $m[2] . $config['suffix']
                            ]
                        ];
                        $shares['/Groupes/@' . $m[2] . '/Travail@' . $m[2]] = [
                            [
                                'share_type' => OCS_GROUP,
                                'permissions' => OCS_PERM_RO,
                                'share_with' => $m[1] . "_" . $m[2] . $config['suffix']
                            ]
                        ];
                        $shares['/Groupes/@' . $m[2] . '/Equipe@' . $m[2]] = [];
                        $shares['/Groupes/@' . $m[2] . '/Echange@' . $m[2]] = [ // [
                                                                                 // 'share_type' => OCS_GROUP,
                                                                                 // 'permissions' => OCS_PERM_RW,
                                                                                 // 'share_with' => $m[1] . "_" . $m[2] . $config['suffix']
                                                                                 // ]
                        ];
                    }

                    if ($m[1] == "Classe" || $m[1] == "Cours") {
                        $i ++;
                        $shares['/Groupes/@' . $m[2] . '/' . $user['cn'] . "@" . $m[2]] = [
                            [
                                'share_type' => OCS_USER,
                                'permissions' => OCS_PERM_RW,
                                'share_with' => $u
                            ]
                        ];
                    }
                }
            }
        }
    }
    return $shares;
}

function form_partages_next($type = "commit")
{
    $types = array(
        "force" => "Forcer le rafraichissement",
        "clean" => "Nettoyer les partages",
        "commit" => "Rafraîchir les partages"
    );
    $html = "Selectionner l'étape suivante :  <SELECT NAME=\"action\" >"; // onchange='this.form.submit()'>";
    foreach ($types as $t => $txt) {
        $html .= "<option value=\"$t\"";
        if ($type == $t) {
            $html .= " selected>";
        } else {
            $html .= ">";
        }
        $html .= "$txt</option>";
    }
    $html .= "<input type=\"submit\" value=\"Valider\">";
    $html .= "</SELECT><br>";
    return $html;
}

/**
 * copie les données d'un utilisateur depuis son partage Samba vers le Cloud
 *
 * @param unknown $config
 * @param unknown $user
 * @param unknown $data
 */
function migrate_to_nc($config, $user, $data, $dest = "")
{
    switch ($data) {
        case 'etab':
            $data = $config['cloud_name'];
        case 'nuage.apps':
            $path = "/home/" . $user['cn'] . "/Docs";
            $dest = $data . ":/Documents";
            $command = "rclone copy " . $path . " " . $dest . " --config=/home/" . $user['cn'] . "/.config/rclone/rclone.conf\n";
            //$path = "/home/" . $user['cn'] . "/photos";
            //$dest = $m[1] . ":/Photos";
            //$command = "rclone copy " . $path . " " . $dest . " -config=/home/" . $user['cn'] . "/.config/rclone.conf";
            break;
        case 'classe':
            $classe = list_classes($config, $user)[0];
            $path = "/var/www/sambaedu/Classes/Classe_" . $classe . "/" . inverse_login($user['cn']);
            $dest = $config['cloud_name'] . ":/Mes travaux/" . $classe;
            $command = "rclone copy " . $path . " " . $dest . " --config=/home/" . $user['cn'] . "/.config/rclone/rclone.conf";
            break;
        default:
            $path = $data;
            $command = "rclone copy " . $path . " " . $dest . " --config=/home/" . $user['cn'] . "/.config/rclone/rclone.conf";
    }
    batch_command($command);
    batch_write();
}

function form_nc_migrate($config, $user, $cloud)
{
    $cl = 0;
    $html = "";
    foreach ($user['memberof'] as $g) {
        $cl += preg_match("/cn=Cloud_.*/i", $g);
    }
    if (! empty($_POST['migrate'])) {
        if (groupaddmember($config, $user['cn'], "Cloud_" . $_POST['migrate'])) {
        migrate_to_nc($config, $user, $_POST['migrate']);
            $html .= "Vos données personnelles sont en train d'être transférées vers le serveur " . $_POST['migrate'] . "<br>";
        } else {
            $html .= "Erreur, transfert vers " . $_POST['migrate'] . " impossible.<br>";
        }
    }
    if (! empty($cloud) && $cl == 0) {
        $html .= '<center>Choisir le serveur vers lequel vous voulez migrer vos données personnelles<br>
<form name="cloud-migrate" action="" method="post">';
        foreach ($cloud as $c) {
            $html .= '<input type=submit name="migrate" value="' . $c . '">';
        }
        $html .= '</form></center>';
    }
    return $html;
}
?>
